﻿// 出力所要時間をファイルに書き出す。
//#define OUTPUT_WORK_TIME

// イメージ出力に以前の処理方法を用いる(Bitmap を作成する)。
//#define USE_OLD_GETOUTPUTIMAGE

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using NiVE2.Plugin.Interface;
using NiVE2.Drawing;
using NiVE2.Utils;
using NiVE2.Utils.Cache;
using NiVE2.Utils.Threading;
using ruche.aviutl.plugins;

namespace ruche.nive2.aviutlPluginPorts
{
    /// <summary>
    /// AviUtl出力プラグインを利用するNiVE出力プラグインクラス。
    /// </summary>
    public class MediaOutput : OutputBase, IOutputInfo
    {
        /// <summary>
        /// 最後に読み込んだAviUtl出力プラグイン情報のキー名。
        /// </summary>
        public static readonly string LastAviUtlOutputPluginKeyName =
            typeof(MediaOutput).Name + ":LastAviUtlOutputPlugin";

        /// <summary>
        /// 対象コンポジション。
        /// </summary>
        private IComposition _compo = null;

        /// <summary>
        /// コンポジションのフレームレート。
        /// </summary>
        private Fraction _frameRate;

        /// <summary>
        /// 出力プラグイン。
        /// </summary>
        private OutputPlugin _plugin = null;

        /// <summary>
        /// 出力形式。
        /// </summary>
        private string _correspondType = null;

        /// <summary>
        /// 既定の拡張子。
        /// </summary>
        private string _defaultExt = null;

        /// <summary>
        /// アボートフラグ。
        /// </summary>
        private bool _aborted = false;

        /// <summary>
        /// 進捗通知デリゲート。
        /// </summary>
        private FrameProgressDelegate _frameProgress = null;

        /// <summary>
        /// 最後に書き出したイメージデータ。
        /// </summary>
        private byte[] _lastImageData = null;

#if USE_OLD_GETOUTPUTIMAGE
        /// <summary>
        /// 最後に書き出した24ビットイメージ。
        /// </summary>
        private Bitmap _lastImage = null;
#endif // USE_OLD_GETOUTPUTIMAGE

        /// <summary>
        /// 最後に書き出したオーディオデータ。
        /// </summary>
        private byte[] _lastAudioData = null;

        /// <summary>
        /// コンストラクタ。
        /// </summary>
        public MediaOutput()
        {
        }

        /// <summary>
        /// デストラクタ。
        /// </summary>
        ~MediaOutput()
        {
            Dispose();
        }

        /// <summary>
        /// 出力プラグインをロードする。
        /// </summary>
        /// <param name="auoFilePath">ファイルのパス。</param>
        private OutputPlugin LoadOutputPlugin(string auoFilePath)
        {
            // チェック
            if (auoFilePath == null)
            {
                throw new ArgumentNullException("auoFilePath");
            }
            if (!File.Exists(auoFilePath))
            {
                throw new ArgumentException(
                    "指定した出力プラグインファイルは存在しません。",
                    "auoFilePath");
            }

            OutputPlugin plugin = null;
            try
            {
                // カレントディレクトリを auoFilePath の場所にする
                CurrentDirectoryStack.Instance.Push(
                    Path.GetDirectoryName(auoFilePath));

                // 出力プラグインインタフェース作成
                plugin = new OutputPlugin(auoFilePath);
            }
            finally
            {
                CurrentDirectoryStack.Instance.Pop();
            }

            return plugin;
        }

        /// <summary>
        /// 出力プラグインを解放する。
        /// </summary>
        private void FreeOutputPlugin()
        {
            if (_plugin != null)
            {
                try
                {
                    // カレントディレクトリをプラグインの場所にする
                    CurrentDirectoryStack.Instance.Push(
                        Path.GetDirectoryName(_plugin.PluginFilePath));

                    _plugin.Dispose();
                    _plugin = null;
                }
                finally
                {
                    CurrentDirectoryStack.Instance.Pop();
                }
            }
        }

        /// <summary>
        /// プラグインが変更されたときの更新処理を行う。
        /// </summary>
        private void UpdatePluginInfo()
        {
            if (_plugin != null)
            {
                // 最後に読み取ったライブラリ情報更新
                LibraryConfig.Datas[LastAviUtlOutputPluginKeyName] =
                    _plugin.PluginFilePath;
                try
                {
                    LibraryConfig.Datas.Save();
                }
                catch { }

                // ファイルフィルタと既定の拡張子設定
                _correspondType = _plugin.FileFilter;
                _defaultExt = "";
                int pos = _correspondType.IndexOf('|');
                if (pos >= 0)
                {
                    pos = _correspondType.IndexOf('.', pos + 1);
                    if (pos >= 0)
                    {
                        int last = _correspondType.IndexOfAny(
                            new char[] { ';', '|' }, pos + 1);
                        if (last < 0) { last = _correspondType.Length; }
                        string ext =
                            _correspondType.Substring(pos, last - pos);
                        if (ext.IndexOfAny(Path.GetInvalidFileNameChars()) < 0)
                        {
                            _defaultExt = ext;
                        }
                    }
                }
            }
            else
            {
                _correspondType = "メディアファイル (*.*)|*.*";
                _defaultExt = "";
            }
        }

        /// <summary>
        /// 終了処理を行う。
        /// </summary>
        public void Finish()
        {
            try
            {
                // プラグイン解放
                FreeOutputPlugin();

                // バッファ解放
#if USE_OLD_GETOUTPUTIMAGE
                if (_lastImage != null)
                {
                    _lastImage.Dispose();
                    _lastImage = null;
                }
#endif // USE_OLD_GETOUTPUTIMAGE
                _lastImageData = null;
                _lastAudioData = null;
            }
            catch { }
        }

        /// <summary>
        /// 進捗通知イベント処理を行う。
        /// </summary>
        /// <param name="frames">処理済みフレーム数。</param>
        /// <param name="totalFrames">総フレーム数。</param>
        /// <returns></returns>
        private bool UpdateFrameProgress(int frames, int totalFrames)
        {
            // 時間進捗算出
            // マルチパスの場合もあるので、進捗割合を合計時間に乗算するようにする
            double pastTime =
                (this.EndTime - this.StartTime) * frames / totalFrames;

            // NiVE2に通知
            _aborted = OnRenderingProgress(this.StartTime + pastTime);

            return true;
        }

        #region OutputBase メンバ

        public override bool CanVideo
        {
            get { return ((OutputTypes & OutputFlags.Video) != 0); }
        }

        public override bool CanAudio
        {
            get { return ((OutputTypes & OutputFlags.Audio) != 0); }
        }

        public override bool HaveSettingWindow
        {
            get { return true; }
        }

        public override string CorrespondingType
        {
            get { return _correspondType; }
        }

        public override string DefaultExtension
        {
            get { return _defaultExt; }
        }

        public override void Initialize(IComposition composition)
        {
            // ライブラリ設定のロード
            LibraryConfig.Datas.Load();

            // コンポジション状態保存
            _compo = composition;
            _frameRate = new Fraction(_compo.Setting.FrameRate);

            // デリゲート初期化
            this.FrameProgress =
                new FrameProgressDelegate(UpdateFrameProgress);

            // 前回最後に読み込んだプラグインの読み込み試行
            if (LibraryConfig.Datas.ContainsKey(LastAviUtlOutputPluginKeyName))
            {
                try
                {
                    string auoPath =
                        LibraryConfig.Datas[LastAviUtlOutputPluginKeyName]
                            as string;
                    if (auoPath != null)
                    {
                        // プラグインをロード
                        _plugin = LoadOutputPlugin(auoPath);
                    }
                }
                catch { }
            }

            // プラグイン情報初期化
            UpdatePluginInfo();

            // ファイル名に既定の拡張子を付ける
            this.FileName =
                Path.ChangeExtension(this.FileName, this.DefaultExtension);

            // その他の初期化
            _aborted = false;
        }

        public override void Setting(IntPtr hwnd)
        {
            using (MediaOutputConfigDialog dlg = new MediaOutputConfigDialog())
            {
                // パラメータ初期化
                dlg.Plugin = _plugin;

                // ダイアログ表示
                if (dlg.ShowDialog(Form.FromHandle(hwnd)) == DialogResult.OK)
                {
                    // 設定取得
                    if (dlg.Plugin != _plugin)
                    {
                        // プラグイン変更
                        FreeOutputPlugin();
                        _plugin = dlg.Plugin;

                        // プラグイン情報更新
                        UpdatePluginInfo();

                        // 出力ファイルの拡張子変更
                        if (DefaultExtension != "")
                        {
                            this.FileName = Path.ChangeExtension(
                                this.FileName, DefaultExtension);
                        }
                    }
                }
            }
        }

        public override void SetSetting(OutputSetting setting)
        {
            this.FileName = setting.FileName;
            this.StartTime = setting.StartTime;
            this.EndTime = setting.EndTime;
        }

        public override OutputSetting GetSetting()
        {
            return new OutputSetting(
                this.PluginName, this.FileName, this.StartTime, this.EndTime);
        }

        public override void StartRendering()
        {
            bool result = false;
            bool curDirPushed = false;

            try
            {
                // プラグインがロードされていないならば出力不可
                if (_plugin == null) { return; }

                // カレントディレクトリをプラグインの場所にする
                CurrentDirectoryStack.Instance.Push(
                    Path.GetDirectoryName(_plugin.PluginFilePath));
                curDirPushed = true;

#if OUTPUT_WORK_TIME
                // 処理開始時間
                DateTime startTime = DateTime.Now;
#endif // OUTPUT_WORK_TIME

                // 出力処理
                _plugin.Execute(this);
                result = true;

#if OUTPUT_WORK_TIME
                // 計測した合計時間をファイルに書き出す
                TimeSpan span = DateTime.Now - startTime;
                File.AppendAllText(
                    this.FileName + "_worktime.txt",
                    "OUTPUT TIME : " + span.ToString() + "\n",
                    Encoding.Default);
#endif // OUTPUT_WORK_TIME
            }
            catch (Exception ex)
            {
#if DEBUG
                throw ex;
#else
                MessageBox.Show(
                    ex.Message,
                    "エラー",
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);

                result = false;
#endif
            }
            finally
            {
                if (curDirPushed)
                {
                    CurrentDirectoryStack.Instance.Pop();
                }

                // 結果を通知する
                OnRenderingComplete(this.IsOutputAborted, !result);
            }
        }

        #endregion

        #region PluginBase メンバ

        public override string PluginName
        {
            get { return "AviUtl出力プラグイン利用"; }
        }

        public override string Author
        {
            get { return "ルーチェ"; }
        }

        public override string InfoLink
        {
            get { return "http://www.ruche-home.net/"; }
        }

        public override string Description
        {
            get
            {
                return "AviUtlの出力プラグイン(*.auo)を利用してメディアファイルを書き出します。";
            }
        }

        #endregion

        #region IOutputInfo メンバ

        public OutputFlags OutputTypes
        {
            get
            {
                return (_plugin == null) ? 0 : _plugin.OutputTypes;
            }
        }

        public int OutputImageWidth
        {
            get { return _compo.Size.Width; }
        }

        public int OutputImageHeight
        {
            get { return _compo.Size.Height; }
        }

        public int OutputImageFrameRateBase
        {
            get { return _frameRate.Numerator; }
        }

        public int OutputImageFrameRateScale
        {
            get { return _frameRate.Denominator; }
        }

        public int OutputImageFrameCount
        {
            get
            {
                double length = this.EndTime - this.StartTime;
                return (int)(
                    length * OutputImageFrameRateBase / OutputImageFrameRateScale);
            }
        }

        public int OutputImageFrameSize
        {
            get
            {
                return (((OutputImageWidth * 3 + 3) & ~3) * OutputImageHeight);
            }
        }

        public int OutputAudioSampleRate
        {
            get { return (int)ReferenceData.WaveFormat.nSamplesPerSec; }
        }

        public int OutputAudioChannel
        {
            get { return ReferenceData.WaveFormat.nChannels; }
        }

        public int OutputAudioSampleCount
        {
            get
            {
                double length = this.EndTime - this.StartTime;
                return (int)(length * OutputAudioSampleRate);
            }
        }

        public int OutputAudioSampleSize
        {
            get { return ReferenceData.WaveFormat.nBlockAlign; }
        }

        public string OutputFilePath
        {
            get { return this.FileName; }
        }

        public bool IsOutputAborted
        {
            get { return _aborted; }
        }

        public byte[] GetOutputImage(int frame)
        {
            long cacheSize32bpp = OutputImageWidth * OutputImageHeight * 4;

            // バッファ確保
            if (
                _lastImageData == null ||
                _lastImageData.Length < OutputImageFrameSize)
            {
                CacheManager.Manager.DemandTemporalMemory(OutputImageFrameSize);
                _lastImageData = new byte[OutputImageFrameSize];
                CacheManager.Manager.AddWeakCache(_lastImageData, _lastImageData.Length);
            }
#if USE_OLD_GETOUTPUTIMAGE
            if (
                _lastImage == null                      ||
                _lastImage.Width  != OutputImageWidth   ||
                _lastImage.Height != OutputImageHeight)
            {
                if (_lastImage != null) { _lastImage.Dispose(); }

                CacheManager.Manager.DemandTemporalMemory(OutputImageFrameSize);
                _lastImage = new Bitmap(
                    OutputImageWidth, OutputImageHeight, PixelFormat.Format24bppRgb);
                CacheManager.Manager.AddWeakCache(_lastImage, OutputImageFrameSize);
            }
#endif // USE_OLD_GETOUTPUTIMAGE

            // 対象時間算出
            double time =
                (double)frame * OutputImageFrameRateScale /
                OutputImageFrameRateBase;
            time += this.StartTime;

#if USE_OLD_GETOUTPUTIMAGE

            // 対象画像を24ビットBMPに描画
            using (CacheLocker cacheLock = new CacheLocker(cacheSize32bpp * 2))
            using (NBitmap target = _compo.RenderingImage(time))
            using (Bitmap orgBmp = target.ToBitmap())
            using (Graphics g = Graphics.FromImage(_lastImage))
            {
                g.CompositingMode = CompositingMode.SourceCopy;
                g.DrawImage(orgBmp, Point.Empty);
            }

            // データを配列へ書き出す
            BitmapData bmpData = _lastImage.LockBits(
                new Rectangle(0, 0, OutputImageWidth, OutputImageHeight),
                ImageLockMode.ReadOnly, PixelFormat.Format24bppRgb);
            Marshal.Copy(
                bmpData.Scan0, _lastImageData, 0, OutputImageFrameSize);
            _lastImage.UnlockBits(bmpData);
            bmpData = null;

#else // USE_OLD_GETOUTPUTIMAGE

            // 32,24ビットBMPの行バイト数算出
            int lenX32 = OutputImageWidth * 4;
            int lenX24 = (OutputImageWidth * 3 + 3) & ~3;

            // 24ビットBMPの行余白バイト数算出
            int lenX24Ext = lenX24 - (OutputImageWidth * 3);

            // 対象画像を取得
            using (CacheLocker cacheLock = new CacheLocker(cacheSize32bpp))
            using (NBitmap target = _compo.RenderingImage(time))
            {
                // 対象画像を24ビットBMP配列へ書き出す
                byte[] data = target.GetData();
                Parallel.For(0, target.Height, (y) =>
                {
                    int sx = lenX32 * y;
                    int sxEnd = sx + lenX32;
                    int dx = lenX24 * y;
                    for (; sx < sxEnd; sx += 4, dx += 3)
                    {
                        // α値を考慮する
                        _lastImageData[dx] =
                            (byte)(((uint)data[sx] * data[sx + 3] + 127) / 255);
                        _lastImageData[dx + 1] =
                            (byte)(((uint)data[sx + 1] * data[sx + 3] + 127) / 255);
                        _lastImageData[dx + 2] =
                            (byte)(((uint)data[sx + 2] * data[sx + 3] + 127) / 255);
                    }
                    Array.Clear(_lastImageData, dx, lenX24Ext);
                });
            }

#endif // USE_OLD_GETOUTPUTIMAGE

            return _lastImageData;
        }

        public byte[] GetOutputAudio(int start, int length)
        {
            // 対象時間算出
            double startTime = (double)start / OutputAudioSampleRate;
            double lastTime = (double)(start + length) / OutputAudioSampleRate;
            double span = lastTime - startTime;
            startTime += this.StartTime;

            // データ取得
            CacheManager.Manager.DemandTemporalMemory(
                (length + 1) * ReferenceData.WaveFormat.nBlockAlign);
            _lastAudioData = _compo.RenderingAudio(startTime, span);
            CacheManager.Manager.AddWeakCache(
                _lastAudioData, _lastAudioData.Length);

            return _lastAudioData;
        }

        public OutputVideoFrameFlags GetVideoFrameFlags(int frame)
        {
            // フレーム情報はわからないので None を返す
            return OutputVideoFrameFlags.None;
        }

        public FrameProgressDelegate FrameProgress
        {
            get { return _frameProgress; }
            set { _frameProgress = value; }
        }

        public UpdatePreviewDelegate UpdatePreview
        {
            // NiVE2では使わない
            get { return null; }
            set { ; }
        }

        #endregion

        #region IDisposable メンバ

        public override void Dispose()
        {
            Finish();
        }

        #endregion
    }
}
